import numpy as np
import math

from astropy import units as ut
from matplotlib import pyplot as plt
from matplotlib.offsetbox import AnchoredText
from os import path, makedirs
import base64
import io

# -- below are slight modifications of
# original author: Min-Su Shin , University of Michigan ----- #-#

def linear(inputArray, scale_min=None, scale_max=None):
    """
Performs linear scaling of the input numpy array.

  :inputArray:
    numpy array
    image data array

  :scale_min:
    float
    minimum data value
  :scale_max:
    float
    maximum data value
  :return:
    numpy array
    image data array

	"""
    if not inputArray.max==0:
      imageData=np.array(inputArray, copy=True)

      if scale_min == None:
          scale_min = imageData.min()
      if scale_max == None:
          scale_max = imageData.max()

      imageData = imageData.clip(min=scale_min, max=scale_max)
      imageData = (imageData -scale_min) / (scale_max - scale_min)
      indices = np.where(imageData < 0)
      imageData[indices] = 0.0
      indices = np.where(imageData > 1)
      imageData[indices] = 1.0

    return imageData

def sqrt(inputArray, scale_min=None, scale_max=None):
    """
Performs sqrt scaling of the input numpy array.

  :inputArray:
    numpy array
    image data array

  :scale_min:
    float
    minimum data value
  :scale_max:
    float
    maximum data value
  :return:
    numpy array
    image data array


	"""
    if not inputArray.max == 0:
        imageData=np.array(inputArray, copy=True)

    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()

    imageData = imageData.clip(min=scale_min, max=scale_max)
    imageData = imageData - scale_min
    indices = np.where(imageData < 0)
    imageData[indices] = 0.0
    imageData = np.sqrt(imageData)
    imageData = imageData / math.sqrt(scale_max - scale_min)

    return imageData


def log(inputArray, scale_min=None, scale_max=None,factor=2.0):
    """
Performs log10 scaling of the input numpy array.

  :inputArray:
    numpy array
    image data array

  :scale_min:
    float
    minimum data value

  :scale_max:
    float
    maximum data value

  :factor:
    float
    logarithm multiplication factor.

  :return:
    numpy array
    image data array
	"""
    if not inputArray.max == 0:
        imageData = np.array(inputArray, copy=True)


    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()
    #factor =math.log10(scale_max - scale_min)
    indices0 = np.where(imageData < 0)
    indices1 = np.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = np.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    try:
        imageData[indices1] = np.log10(imageData[indices1])/factor
    except:
        print ("Error on math.log10 ")
    return imageData

# ---------- author : @avialxee ---------#-#


def save_fig(plt, fig, kind='base64', output='output.jpg'):
    
    if kind == 'base64':
        buf = io.BytesIO()
        fig.savefig(buf, format='png', bbox_inches='tight',
                    transparent=True, pad_inches=0)
        buf.seek(0)
        string = base64.b64encode(buf.read())
        plt.close()
        return string
    elif kind == 'plot':
        plt.show()
        return 'plotted'
    else :
        if not path.exists('output'):
            makedirs('output')
        newPath = 'output/'+output
        opt = newPath
        if path.exists(newPath):
            numb = 1
            while path.exists(newPath):
                newPath = "{0}_{2}{1}".format(
                    *path.splitext(opt) + (numb,))
                try :
                    if path.exists(newPath):
                        numb += 1 
                except:
                    pass               
        fig.savefig(newPath, format=kind, bbox_inches='tight',
                    pad_inches=0)
        print("saved {}".format(newPath))
        plt.close()
        return newPath


def overlayc (r,g,b,c,lvl,cmin) :
  """
**Returns RGB stacked image and contour levels to be overlayed.**


*Input*

  :r/g/b:
    (2-D array)

    survey 2-D array input data

  :c:
    (2-D array)

    survey data for contour

  :lvl:
    (int)

    number of levels to draw

  :cmin:
    (float)
    minimum value selected for the contour

*Returns*

  :Img:
    (np array)
    dimension: (px,px,3)

  :lvlc:
    (list)
    dimension: None or of length len(lvl)
  """
  ri = normals(r)
  gi = normals(g)
  bi = normals(b)
  if ri.max() != 0 and ri.min()!=0 :
    ri = sqrt(ri,scale_min=0.1*np.std(ri),scale_max=np.max(ri))
  if gi.max() != 0 and gi.min()!=0 :
    gi = sqrt(gi, scale_min=0.1*np.std(gi), scale_max=np.max(gi))
  if bi.max() != 0 and bi.min()!=0 :
    bi = sqrt(bi,scale_min=0.1*np.std(bi),scale_max=np.max(bi))
  replace = bi > 1
  bi[replace] = 1
  replace = gi > 1
  gi[replace] = 1
  replace = ri > 1
  ri[replace] = 1

  # Stacking the Images and set up as 8-bit integers (highest value 2^8 = 256)
  img = (np.dstack((ri,gi,bi))*255.99).astype(np.uint8)
  if c.max() > cmin :
    lvlc = np.arange(cmin, c.max(),((c.max() - cmin)/lvl))
  else :
    lvlc = None
  return img, lvlc

def normals(o) :
  """
normalizing before scaling

*Input*

  :o:
    (2D array)
    input survey

*Returns*

  :(2D array):
    dimension same as input survey

  """
  X_scaled =(o - np.median(o))
  if o.max()!=0 :
    X = X_scaled/o.max()
    return X
  else :
    return X_scaled

def pl_powerlawsi(S,S_e,freq= [150, 1420], kind='plot', label="output"):
  """
plot powerlaw of spectralindex

*Inputs*

  :S:
    Total flux of TGSS and NVSS resp (list)
  :S_e:
    error on Total flux of TGSS and NVSS resp (list)
  :freq:
    frequency of TGSS and NVSS resp (list) (default = [150, 1420] )
  :kind:
    (str) choose from 'base64', 'plot', 'jpg','png'
      - 'base64' : base64 encoded
      - 'plot' : outputs plot in window or in your jupyter notebook
      - 'jpg' : saves jpg image output in output/
      - 'png' : saves png image output in output/

*Example*

.. code-block:: python

  >>> S_e = [42.6, 2.6]
  >>> S = [424.0, 60.0]
  >>> pl_si(S,S_e)
  """

  tgss = [float(S[0])]
  nvss = [float(S[1])]
  factor = freq[0]/freq[1]
  si = np.round(np.log(np.divide(tgss, nvss)+1E-5)/np.log(factor+1E-5), 3)
  fig = plt.figure(figsize=(10, 5))
  ax1 = fig.add_subplot(1, 2, 1)
  ax1.errorbar(freq, S, yerr=S_e,
            label ='Line1', color='red')
  ax1.set_xscale('log')
  ax1.set_yscale('log')
  ax1.set_ylabel('Spectral flux (mJy)')
  ax1.set_xticks([freq[0], freq[1]*0.4,freq[1]])
  #ax1.set_yticks([S[0], S[1]*0.4, S[1]])
  ax1.set_xlabel('frequency (MHz)')
  #slope = (S[1]-S[0])/(freq[1]-freq[0])
  anchored_text = AnchoredText(f'spectral index: {si[0]}', loc=1)
  ax1.add_artist(anchored_text)
  
  save_fig(plt,fig,kind,output=label)
  
def pl_RGB(ax, img,title,name,annot=True):
  """
    *Inputs*
    
        :rows:
          (int)
          Total number of rows.
        :columns:
          (int)
          Total number of columns.
        :i:
          (int)
          current number of the cell.
        :wcs:
          (astropy wcs)
          world coordinate system fetched from header of fits
        :img:
          (np array)
          dimension = (px,px,3)
        :fig:
          <maplotlib.pyplot.figure>
        :name:
          (string)
          input name to be show on figure

    *Returns*
    
        adds plot to the figure.
  """
  #--- RGBC plot -------------------#--#

  ax.axis('off')
  ax.imshow(img, origin='lower', cmap='gist_gray')
  if annot:
      ax.annotate("#RADatHomeIndia",(10,10),color='white')
      ax.annotate("By " + str(name),(400-5*len(name),10),color='white')
      ax.set_title('{}'.format(title),
                    y=1, pad=-16, color="white")
  ax.set_autoscale_on(False)

def overlayo(ri, gi, bi, kind = 'IOU'):
  """
  Returns RGB stacked image.

  *Input*

    :ri/gi/bi:
      (2-D array)
      survey 2-D array data in
    :kind:
      (string)
      either IOU or Optical

  *Returns*

    :Img:
      (nd array)
      dimension: (px,px,3)

    :scaling:
        - IOU
          - sqrt(w22), sqrt(dss2r), log(gnuv)(5 to 100% & factor =3.15)
        - Optical
          - sqrt, sqrt, sqrt (min to max)
  """
  if kind == 'IOU':
    ri = sqrt(ri, scale_min=np.percentile(np.unique(ri),1.), scale_max=np.percentile(np.unique(ri),100.))
    gi = sqrt(gi, scale_min=np.percentile(np.unique(gi),1.), scale_max=np.percentile(np.unique(gi),100.))
    #gi = log(gi, scale_min=np.percentile(np.unique(gi),0.), scale_max=np.percentile(np.unique(gi),100.),factor=2.85)
    bi = log(bi, scale_min=np.percentile(np.unique(bi), 5.),
             scale_max=np.percentile(np.unique(bi), 100.), factor=3.15)
    mul_factor = 255/ri.max()
    img = (np.transpose([(ri*mul_factor).astype(np.uint8),(gi*255/gi.max()).astype(np.uint8),(bi*255).astype(np.uint8)], (1, 2, 0)))

  if kind == 'Optical':
    #ri = log(ri, scale_min=np.percentile(np.unique(ri),1.), scale_max=np.percentile(np.unique(ri),100.),factor=0.3)
    ri = sqrt(ri, scale_min=np.min(ri), scale_max=np.percentile(np.unique(ri),100.))
    gi = sqrt(gi, scale_min=1.15*np.min(gi), scale_max=np.percentile(np.unique(gi),100.))
    #gi = log(gi, scale_min=np.percentile(np.unique(gi),0.), scale_max=np.percentile(np.unique(gi),100.),factor=7.85)
    #bi = log(bi, scale_min=np.percentile(np.unique(bi),1.), scale_max=np.percentile(np.unique(bi),100.),factor=3.15)
    bi = sqrt(bi, scale_min=np.min(bi), scale_max=np.percentile(np.unique(bi),100.))
    mul_factor = 255/ri.max()
    img = (np.transpose([(ri*mul_factor).astype(np.uint8),(gi*255.99).astype(np.uint8),(bi*256).astype(np.uint8)], (1, 2, 0)))
  #img = (np.dstack((ri,gi,bi))*255.99).astype(np.uint8)
  #img = np.stack([ri,gi,bi], axis=2)
  return img

def pl_RGBC(rows,columns,i,wcs,svy,lvlc,img,fig,name, pkind='ror',annot=True) :
    """
    *Inputs*

        :rows:
          (int)
          Total number of rows.
        :columns:
          (int)
          Total number of columns.
        :i:
          (int)
          current number of the cell.
        :wcs:
          (astropy wcs)
          world coordinate system fetched from header of fits
        :svy:
          (np array)
          2-D array survey for contours.
        :lvlc:
          (list)
          list of float
        :img:
          (np array)
          dimension = (px,px,3)
        :fig:
          <maplotlib.pyplot.figure>
        :name:
          (string)
          input name to be show on figure
        :pkind:
          (str)
          ror/iou;
          default='ror'

    *Returns*
    
        plots the ror/iou figure.
    """
    ax = fig.add_subplot(rows, columns, i, projection=wcs)
    ax.axis( 'off')
    ax.imshow(img, origin='lower', interpolation='nearest')
    if annot:
      ax.annotate("#RADatHomeIndia",(10,10),color='white')
      ax.annotate("By " + str(name),((400-5*len(name)),10),color='white')
    ax.set_autoscale_on(False)

    plt.contour(svy, lvlc, colors='white')
    if pkind == 'ror' and annot:
      if i==1 :
          ax.set_title("ROR-RGB-C: TGSS(GMRT)-DSS2-NVSS(VLA)-NVSS",
                      y=1, pad=-16, color="white")
          #ax.set_title("")
      if i==2 :
          ax.set_title("ROR-RGB-C: TGSS(GMRT)-DSS2-NVSS(VLA)-TGSS",
                      y=1, pad=-16, color="white")
    elif pkind == 'iou' and annot:
      if i==1 :
          ax.set_title("IOU-RGB-C: WISE(22)-DSS2(red)-GALEX(NUV)-TGSS",
                      y=1, pad=-16, color="white")
      if i==2 :
          ax.set_title("Optical-RGB-C: DSS2(IR)-DSS2(Red)-DSS2(blue)-TGSS",
                      y=1, pad=-16, color="white")


# ------------ arcsec to pixel conversion ----------------------------#-#

def to_pixel(unit_inarcsec, r, px = 480):
    """
Convert arcsec into pixel.
    
    :input:
      unit_inarcsec
        input magnitude (unitless) data in arcsec
      r
        length of square image in deg
      px
        total pixels in one length
    :Returns:
        input data converted to be used onto pixel length rounded upto 2 decimal places.
    """
    nvss_px_scale = px/(r)
    return np.round(((unit_inarcsec/3600) * ut.deg * nvss_px_scale), 2).value


from setuptools import setup

with open("README.md", "r") as rdme:
    desc = rdme.read()

setup(
    name = 'rgbmaker',
    version = '0.0.5.0.4',
    url='https://github.com/avialxee/rgbmaker',
    author='Avinash Kumar',
    author_email='avialxee@gmail.com',
    description='A python package which communicates to different astronomical services and fetches fits and numerical data.',
    py_modules = ["rgbmaker"],
    package_dir={'':'src'},
    classifiers=["Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "License :: OSI Approved :: BSD License",
                 "Intended Audience :: Science/Research",
                 ],
    long_description=desc,
    long_description_content_type = "text/markdown",
    install_requires=["astropy>=4.2.1", "matplotlib>= 3.4.2",
                      "regions", "numpy>= 1.20.3", "astroquery","requests"
                      ],
    extras_require = {
        "dev" : ["pytest>=3.7",
        ]
    },
     entry_points={ 
        "console_scripts": [
            "rgbmaker=rgbmaker.cli:cli",
        ],
    },

)



#-- author: @avialxee ---#--#
import requests
from astropy import coordinates, units as ut
from astropy.wcs import WCS
from astroquery.skyview import SkyView as skv

import numpy as np
from time import perf_counter

from warnings import simplefilter
from astroquery.nvas import Nvas
from astroquery.vizier import Vizier

class RGBMaker:

    def __init__(self, name="", position="", radius=float(0.12), archives=1, imagesopt=2, px=480):
        """
        creating instance for variables to be used throughout the process.
        """

        #--- input variables -----#-#

        self.name = str(name)
        self.position = str(position)
        self.radius = float(radius)
        self.archives = int(archives)
        self.imagesopt = int(imagesopt)

        #--- helper & output variables -----#-#

        self.i = 0
        self.uri = []
        self.info = ""
        self.imglt = []
        self.wcs = None
        self.status = "info"
        self.otext = []
        self.px = px
        self.name, self.c, self.r = self._inp_sanitize()
        self.server_down = False

    def throw_output(self):
        """
        returns output showing status of output. 
        
        TODO 
        ----
        1. Http Status-code-like json output. 200, 202, 404 etc
        2. Show vizier table. ex. otext['vizier]
        """
        return self.status, self.uri, self.info, self.otext

    def submit_query(self):
        """
        takes input of name, position, radius, choice of image and arhives 
        to fetch FITS from Skyview and NVAS.        
        """
        ## ------------- Settings ---------------------
        np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
        #VerifyWarning by astropy for NVSS in imagesopt=2 
        simplefilter('ignore', category=UserWarning)
        _input_svys, _input_sampler = self._sanitize_rgb()
        self.name, self.c, self.r = self._inp_sanitize()
        if self.c is not None :
            self._getNVAS(self.archives)
            try:
                hdu_d_list, error = self.getdd(_input_svys, _input_sampler)
            except:
                if self.c :
                    if self.imagesopt == 1 or self.imagesopt == 2 or self.imagesopt == 3:
                        self.status, self.info = 'info', 'error fetching data from skyview'
                    else:
                        self.status, self.info = 'info', 'No images to return'
                return self.throw_output()
            return self._arr_rgb(_input_svys, hdu_d_list)
        else :
            self.status, self.info = 'warning', 'Please check Coordinates'
            return self.throw_output()
        
        #return hdu_d_list #DEBUG

    def _arr_rgb(self, input_svys, hdu_d_list):
        """
        @input 
            of svy names, data returned for surveys 
        @returns: 
            list containing data, range that can be used to plot in matplotlib.
        """
        _svys_res = None
        if hdu_d_list is not None:
            for result in hdu_d_list:
                if not self.wcs:
                    self.wcs = WCS(result[1])
            _key_svy_dict = list(self.svy_dict('*').keys())
            _val_svy_dict = list(self.svy_dict('*').values())
            _svys_res = {}
            _start = perf_counter()
            if self.wcs:
                for i in range(len(input_svys)):
                    _ind_rgb_dict = _val_svy_dict.index(input_svys[i])
                    _svy_sname = _key_svy_dict[_ind_rgb_dict]
                    _svys_res[_svy_sname] = {'data': hdu_d_list[i][0],
                                            'range': 'value ranges from {:0.4f} to {:0.4f}'.format(hdu_d_list[i][0].min(), hdu_d_list[i][0].max())
                                            }
                    if _svy_sname == 'tgss' or _svy_sname == 'nvss' or _svy_sname=='first':
                        self.otext.append(
                            {input_svys[i]  :  _svys_res[_svy_sname]['range']})
                
            else:
                self.info = "No images found."
                _time_taken = perf_counter()-_start
                self.info += ' Time taken:' + str(np.round(_time_taken, 3))+". "
                self.status = "warning"
                return self.throw_output()        
            return _svys_res
        else:
            self.info = "No data fetched from skyview."
            return self.throw_output()

    def _sanitize_rgb(self):
        """
        returns RGB surveys as array dtype=<U13 and sampler as list to be used by getdd().
        """
        _rgbs = []
        _sampler = [None]
        if (self.c is not None) and (self.imagesopt == 1 or self.imagesopt == 2 or self.imagesopt == 3):
            if int(self.imagesopt) == 1 :
                _rgb1 = self.rgb_dict('ror')
                _rgb2 = self.rgb_dict('iou')
                _rgb3 = self.rgb_dict('optical')
                _rgbs = list(set(_rgb1['rgb']) | set(
                    _rgb2['rgb']) | set(_rgb3['rgb']))

            elif int(self.imagesopt) == 2 :
                _rgb1 = self.rgb_dict('single')
                _rgb2 = self.rgb_dict('ror')
                _rgbs = list(set(_rgb1['rgb']) | set(_rgb2['rgb']))

            elif int(self.imagesopt) == 3 :
                _rgb1 = self.rgb_dict('allwise')
                _rgbs = list(set(_rgb1['rgb']))

            _sampler = np.array([None]*len(_rgbs))
            indices = np.where(np.array(_rgbs) == 'NVSS')
            
            if len(indices[0]):
                _sampler[indices] = 'Lanczos3'
        return _rgbs, _sampler

    def getdd(self, input_svys, sampler):
        """
        takes input svys (array) and fetches fits from skyview using astroquery.
        returns [list([hdulist[0].data, hdulist[0].header]), error]
        """
        _imglt = []
        _imgls, _error = self._run_imgl(
            self.c, self.r, input_svys, sampler)
        for i in range(len(_imgls)):
            if _imgls[i] == 0:
                _imglt.insert(
                    i, [np.zeros((int(self.px), int(self.px))), None])
            else:
                _imglt.append([_imgls[i][0][0].data, _imgls[i][0][0].header])
                if not self.wcs:
                    self.wcs = WCS(_imgls[i][0][0].header)
        return _imglt, _error

    def _getNVAS(self, archives):
        archives = self.archives
        if archives and int(archives) == 2 and self.c:
            text=""     
            try :
                nvas_urls = Nvas.get_image_list(self.c,radius=2*ut.arcsec)
                text += " " + str(len(nvas_urls)) + \
                    " Image(s) found in NVAS: "
                i = 1
                for nvas in nvas_urls :
                    if i<=5 :
                        text += " <a href='" + str(nvas) + "' target='_blank' rel='noreferrer noopener'>[" + str(i) + "]</a>"
                        i+=1
                self.otext.append({'NVAS': text})
            except :
                self.otext.append({'NVAS' : 'NVAS server error'})

    def _run_imgl(self, c, r, _in_svys, sam):
        """
        Takes coordinate, radius and surveys to be fetched from 
        skyview and sends to _get_imgl_pool.
        returns a list of hdul and errors requested by getdd.
        """
        result = [0]*len(_in_svys)
        imglt, _error = [None]*2
        _sam = sam
        try:
            #print(_in_svys)
            try:
                for ind in range(len(_in_svys)):
                    imglt = self._get_imgl_pool(
                        [c, _in_svys[ind], r, result, ind,  _sam[ind]])
            except:
                _error = "problem with survey: " + str(_in_svys)
        except Exception as e:
            _error = e
        return imglt, _error

    def _get_imgl_pool(self, cals):
        """
        Requests Fits from Skyview one by one. Written for multithread pooling.
        returns a list of hdul requested by _run_imgl
        """
        c, svy, r, queue, ind, _sam = cals
        if _sam == 'None':
            _sam = None
        try:
            imglr = skv.get_images(position=c, survey=svy, pixels=str(
                self.px), radius=r, scaling="Linear", sampler=_sam, cache='True')
            queue[ind] = imglr
        except requests.exceptions.ConnectionError as e:
            self.server_down = True
            self.info = e
        except Exception as e:
            # --- if file not found/doesn't exist. Program will continue.
            print("{} not found ".format(svy))
            pass
        return queue

    def _inp_sanitize(self):
        """
        sanitizing inputs: name, radius, position
        """

        #------------ name --#
        name = str(self.name)
        if len(name) <= 2:
            name = "Anonymous"
        if len(name) > 27:
            name = name[:28]

        #---------- radius --#
        if float(str(self.radius)) > 2.0:
            self.radius = 2.0
        r = float(str(self.radius))*ut.degree

        #----- coordinates --#
        c = None
        try:
            if not ('' or None) :
                c = coordinates.SkyCoord.from_name(self.position, frame='fk5')
                self.info = "success"
                self.status = "success"
                return name, c, r
        except:
            self.info = " Please check coordinates"
            self.status = "warning"
            #print(self.throw_output())
            return name, c, r

    @classmethod
    def rgb_dict(cls, combination):
        """
        dictionary for creating RGB images and single survey images using svy_dict().
        """
        _sampling_b, _r, _g, _b = [None]*4
        if "iou" in combination:
            _r = cls.svy_dict('w22')
            _g = cls.svy_dict('dss2r')
            _b = cls.svy_dict('gnuv')
        elif "ror" in combination:
            _r = cls.svy_dict('tgss')
            _g = cls.svy_dict('dss2r')
            _b = cls.svy_dict('nvss')
            _sampling_b = 'Lanczos3'
        elif "optical" in combination:
            _r = cls.svy_dict('dss2ir')
            _g = cls.svy_dict('dss2r')
            _b = cls.svy_dict('dss2b')
        elif "single" in combination:
            _g = cls.svy_dict('dss2r')
            _r = cls.svy_dict('first')
            return {'rgb': [_r,_g], 'sampler': [None]}
        elif "allwise" in combination:
            _r = cls.svy_dict('w22')
            _g = cls.svy_dict('w4_6')
            _b = cls.svy_dict('w3_4')
            return {'rgb': [_r,_g, _b], 'sampler': [None, None, None]}

        return {'rgb': [_r, _g, _b], 'sampler': [None, None, _sampling_b]}
    
    def vz_query(self):
        """
        query from vizier for NVSS and TGSS.
        @returns:
            MajorAxis (unitless) (Vizier : arcsec), 
            MinorAxis (unitless) (Vizier : arcsec),
            PA  (unitless) (Vizier : deg), 
            pixel based position (single arguement)
        """
        _tgss_viz, _nvss_viz = [None]*2
        if self.imagesopt == 2:
            try:
                try: 
                    _tviz = Vizier(columns=self.vzc_dict('tgss')).query_region(
                        self.c, self.r, catalog=self.vzb_dict('tgss'))
                    _tra = _tviz[0]['RAJ2000']  # *ut.deg
                    _tdec = _tviz[0]['DEJ2000']  # *ut.deg
                    _tMaj = _tviz[0]['Maj']
                    _tMin = _tviz[0]['Min']
                    _s_tgss = _tviz[0]['Stotal']
                    _es_tgss = _tviz[0]['e_Stotal']
                    if "PA" in _tviz[0].columns:
                        _tPA = _tviz[0]["PA"]
                    else:
                        _tPA = 0
                    _center_tgss = coordinates.SkyCoord(ra=_tra, dec=_tdec)
                    _center_tgss_px = self.wcs.world_to_pixel(_center_tgss)
                    _tgss_viz = [_tMaj, _tMin, _tPA, _center_tgss_px, _s_tgss, _es_tgss]
                finally:
                    _nviz = Vizier(columns=self.vzc_dict('nvss')).query_region(
                        self.c,self.r, catalog=self.vzb_dict('nvss'))
                    _nra = coordinates.Angle(_nviz[0]['RAJ2000'], unit=ut.hour)  # .deg
                    _ndec = coordinates.Angle(_nviz[0]['DEJ2000'], unit=ut.deg)  # .deg
                    #_nrad = coordinates.Angle(_nviz[0]['RAJ2000'], unit=ut.hour).deg
                    #_ndecd = coordinates.Angle(_nviz[0]['DEJ2000'], unit=ut.deg).deg for scatter plots
                    _nMaj = _nviz[0]["MajAxis"]
                    _nMin = _nviz[0]["MinAxis"]
                    _s_nvss = _nviz[0]['S1.4']
                    _es_nvss = _nviz[0]['e_S1.4']
                    if "PA" in _nviz[0].columns:
                        _nPA = _nviz[0]["PA"]
                    else:
                        _nPA = 0
                    _center_nvss = coordinates.SkyCoord(ra=_nra, dec=_ndec)
                    _center_nvss_px = self.wcs.world_to_pixel(_center_nvss)
                    _nvss_viz = [_nMaj, _nMin, _nPA, _center_nvss_px, _s_nvss, _es_nvss]
            except:
                self.info = " no data in catalog"

        return _tgss_viz, _nvss_viz

    @classmethod
    def svy_dict(cls, svy=None):
        """
        dictionary for surveys to be used to fetch FITS from skyview using astroquery.
        """
        cls.svys = Surveys()
        _dict = cls.svys.__dict__
        if svy in _dict:
            return _dict[svy]
        if svy is None :
            return cls.svys
        elif svy == '*':
            return _dict
        elif svy == 'v':
            return _dict.values()
        elif svy == 'k':
            return _dict.keys()
        else:
            return "'{}' is not a valid survey. please choose one from {} or use any of '*' 'v' 'k' ".format(svy, _dict.keys())

    @staticmethod
    def vzc_dict(svy):
        """
        vizier column dictionary
        dictionary for NVSS, TGSS column using vizier
        TODO: add column for errors too
        """
        columns = {'tgss': ['RAJ2000', 'DEJ2000', 'Maj', 'Min', 'PA', 'Stotal', 'e_Stotal'],
                   'nvss': ['RAJ2000', 'DEJ2000', 'MajAxis', 'MinAxis', 'PA', '+NVSS', ' S1.4', 'e_S1.4']
                   }
        if svy in columns:
            return columns[svy]
        elif svy == '*':
            return columns
        elif svy == 'v':
            return columns.values()
        elif svy == 'k':
            return columns.keys()
        else:
            return "'{}' is not a valid survey. please choose one from {} or use any of '*' 'v' 'k' ".format(svy, columns.keys())

    @staticmethod
    def vzb_dict(svy):
        """
        vizier base dictionary
        dictionary for NVSS, TGSS catalog using vizier
        """
        base = {'tgss': 'J/A+A/598/A78/table3',
                'nvss': 'VIII/65/nvss'
                }
        if svy in base:
            return base[svy]
        elif svy == '*':
            return base
        elif svy == 'v':
            return base.values()
        elif svy == 'k':
            return base.keys()
        else:
            return "'{}' is not a valid survey. please choose one from {} or use any of '*' 'v' 'k' ".format(svy, base.keys())

class Surveys(RGBMaker):
    
    def __init__(self) -> None:

        self.__dict__ = {'tgss': 'TGSS ADR1',
                'nvss': 'NVSS',
                'first': 'VLA FIRST (1.4 GHz)',
                'dss2r': 'DSS2 Red',
                'dss2ir': 'DSS2 IR',
                'dss2b': 'DSS2 Blue',
                'w22': 'WISE 22',
                'w12': 'WISE 12',
                'w4_6': 'WISE 4.6',
                'w3_4': 'WISE 3.4',
                'gnuv': 'GALEX Near UV'}



from rgbmaker import RGBMaker as rgbmaker
from rgbmaker.imgplt import pl_RGB, to_pixel, pl_RGBC, overlayc, overlayo, sqrt, save_fig
from rgbmaker.tgss_spidx import find_spidx

from matplotlib import pyplot as plt
from regions import PixCoord, EllipsePixelRegion
from astropy.coordinates import Angle, SkyCoord
from astropy import units as ut
from matplotlib.collections import PatchCollection

import matplotlib.patches as mpatches
import numpy as np

import urllib

from time import perf_counter

import sys

def query(name="", position="", radius=float(0.12), archives=1, 
                    imagesopt=2, kind='base64', spidx_file=None, px=480, annot=True):
    """
    
.. jupyter-execute::

    from rgbmaker.fetch import query
    query(name='Avi', position='3C 33.1', radius=0.12, kind='plot')

*Parameters*

    :name:
        (Optional) (default=Anonymous) (string)  

        Your name will be displayed on the image enabling mentors, professors, 
        fellow students to be able to recognize your work. Credit is important!

    :position: 
        (Required)

        The object name or the coordinates of the object in the FK5 (J2000) system. 
        Ex: "14 09 48.86 -03 02 32.6", M87, NGC1243, without quotes.

    :radius:
        (Required) (default = 0.12) (float)

        The size of the image in degrees, this size will be used for the 
        `field of view <https://en.wikipedia.org/wiki/Field_of_view/>`_ in the resultant image. 
        For reference, in the night sky, the moon is about 0.52 degrees across.

    :imagesopt:
        (default=2)(string)(values=1,2,3)

        This dropdown gives you a choice of the composite images you want to create. 

        **IOU ROR Optical (option = 1)**
            
            *This option returns four images.*

                1. There are two `ROR <https://radathomeindia.org/rgbmaker-info#what-is-iou-ror-and-rgb>_` 
                *(Radio (TGSS ADR1) - Optical (DSS2Red) - Radio (NVSS))* images. 
                One with TGSS Contours and another with NVSS Contours. 
                
                2. The third image is an `IOU <https://radathomeindia.org/rgbmaker-info#what-is-iou-ror-and-rgb>_` 
                *(Infrared (WISE 22) - Optical (DSS2 Red) - Ultraviolet (Galex Near UV))* with TGSS Contours.

                3. The final RGB image is an optical image with *(DSS2IR - DSS2Red - DSS2Blue)* with TGGSS Contours.

        **Composite Contours on DSS2R  (option = 2)**
        
                *This option returns two images.*

                1. The first is a `ROR <https://radathomeindia.org/rgbmaker-info#what-is-iou-ror-and-rgb>_` Image with TGSS contours. 
                The various symbol seen on the image is the `catalog <https://en.wikipedia.org/wiki/List_of_astronomical_catalogues/>`_ 
                data of the respective survey.

                2. The second image is a composite image with DSS2Red background and contours of various 
                radio surveys like TGSS, NVSS, and VLA First (if available).

    
    :archives:
        (default=1)(string)

        This dropdown currently offers access to the NVAS image archive. Selecting this option will 
        return the top 5 results from NVAS (if exists). These can be downloaded as .imfits files (open with DS9) 
        by using save as an option when right-clicked.
    
    :kind:
        (default='base64')
        choose from base64, plot, png, jpg to show base64 of resultant image, plot on output, save png/jpg files
        
    :spidx_file:
        (Default=None)
        enter path to spidx.fits file that contains spectral index data (see example 1).
    
    :px:
        (default=480)
        change pixel value for the final resulatant image.
    
    :annot:
        (default=True)
        remove any annotation by setting this to False.

    """
    fetch_q = rgbmaker(name=name, position=position,
                  radius=radius, archives=archives, imagesopt=imagesopt)
    name = fetch_q.name
    start = perf_counter()
    fetch_q.px = px
    val = fetch_q.submit_query()
    
    level_contour=4

    if fetch_q.server_down:
        fetch_q.status, fetch_q.info = 'warning', 'SkyView is down!'
        return fetch_q.throw_output()

    else:
        if fetch_q.c:
            fetch_q.otext.append({'Target center': fetch_q.c.to_string('hmsdms')})
        if fetch_q.imagesopt == 1 and fetch_q.c:
            # --- using variables for readability -#--#
            tgss, dss2r, nvss, w22, gnuv, dss2i, dss2b = val['tgss']['data'], val[
                'dss2r']['data'], val['nvss']['data'], val['w22']['data'], val[
                    'gnuv']['data'], val['dss2ir']['data'], val['dss2b']['data']

            # --- creating images ----------#--#
            img1, lvlc1 = overlayc(tgss, dss2r, nvss, nvss, level_contour, 0.0015) # NVSS
            img2, lvlc2 = overlayc(tgss, dss2r, nvss, tgss, level_contour, 0.015) # TGSS
            img3 = overlayo(w22,dss2r,gnuv, kind='IOU')
            img4 = overlayo(dss2i,dss2r,dss2b, kind='Optical')
            if lvlc2 is not None:
                fetch_q.otext.append({'TGSS contour ': (str(np.round(lvlc2, 3)))}) # TGSS
            if lvlc1 is not None:
                fetch_q.otext.append({'NVSS contour ': (str(np.round(lvlc1, 4)))}) # NVSS

            # -------- plotting first plot -------------#--#
            plt.ioff()
            fig = plt.figure(figsize=(20, 20))

            pl_RGBC(1, 2, 1, fetch_q.wcs, val['nvss']['data'], lvlc1, img1, fig, name,annot=annot)
            pl_RGBC(1, 2, 2, fetch_q.wcs, val['tgss']['data'], lvlc2, img2, fig, name,annot=annot)
            plt.subplots_adjust(wspace=0.01, hspace=0.01)

            #-------- Saving first plot ------#--#
            string = save_fig(plt, fig, kind)
            
            fetch_q.uri.append(
                {'img1': 'data:image/png;base64,' + urllib.parse.quote(string)})

            #-------- plotting second plot -----#--#
            plt.ioff()
            fig1 = plt.figure(figsize=(20, 20))
            pl_RGBC(1, 2, 1, fetch_q.wcs, val['tgss']['data'], lvlc2, img3, fig1, name, pkind='iou',annot=annot)
            pl_RGBC(1, 2, 2, fetch_q.wcs,
                val['tgss']['data'], lvlc2, img4, fig1, name, pkind='iou',annot=annot)
            plt.subplots_adjust(wspace=0.01, hspace=0.01)

            #-------- Saving second plot ------#--#
            string1 = save_fig(plt, fig1, kind)
            
            fetch_q.uri.append(
                {'img2': 'data:image/png;base64,' + urllib.parse.quote(string1)})

            #-------- Output for success -----#--#
            time_taken = perf_counter()-start
            fetch_q.info = 'completed in ' + str(np.round(time_taken, 3))+". "
            fetch_q.status = "success"
            
        elif fetch_q.imagesopt == 2 and fetch_q.c:
            
            tgss, dss2r, nvss, first = val['tgss']['data'], val[
                'dss2r']['data'], val['nvss']['data'], val['first']['data']

            # --- plots initialization ------#--#
            #img1, lvlc1 = overlayc(tgss, dss2r, nvss, tgss, level_contour, 0.015)
            if tgss.max() > 0.015:
                lvlct = np.arange(0.015, tgss.max(), ((tgss.max() - 0.015)/level_contour))
                fetch_q.otext.append({'TGSS contour ': (str(np.round(lvlct.tolist(), 3)))})
            else:
                lvlct = None
            if first.max() > 0.0005:
                lvlcf = np.arange(0.0005, first.max(),
                                    ((first.max() - 0.0005)/level_contour))
                fetch_q.otext.append({'FIRST contour ': (str(np.round(lvlcf, 4)))})
            else:
                lvlcf = None
            if nvss.max() > 0.0015:
                lvlcn = np.arange(0.0015, nvss.max(), ((nvss.max() - 0.0015)/level_contour))
                fetch_q.otext.append({'NVSS contour ': (str(np.round(lvlcn.tolist(), 4)))})
            else:
                lvlcn = None

            #--- plotting --------------------#--#
            plt.ioff()
            fig = plt.figure(figsize=(20, 20))
            title = ' TGSS(GMRT)-NVSS(VLA)-DSS2R(DSS)'
            dss2r = sqrt(dss2r, scale_min=np.percentile(
                np.unique(dss2r), 1.), scale_max=np.percentile(np.unique(dss2r), 100.))

            #--- RGBC plot -------------------#--#
            ax1 = fig.add_subplot(1,2,1, projection=fetch_q.wcs) 
            pl_RGB(ax1, dss2r, title, name, annot)
            
            #--- vizier access ---------------#--
            # TODO : return table in output
            tgss_viz, nvss_viz = fetch_q.vz_query()
            if tgss_viz is not None:
                tmaj, tmin, tPA, tcen, s_tgss, es_tgss = tgss_viz
            if nvss_viz is not None:
                nmaj, nmin, nPA, ncen, s_nvss, es_nvss = nvss_viz
            try:
                try:
                    patch1 = []
                    for i in range(len(tcen[0])):
                        x, y = tcen[0][i], tcen[1][i]
                        ce = PixCoord(x, y)
                        a = to_pixel(tmaj[i], fetch_q.r, px)
                        b = to_pixel(tmin[i], fetch_q.r, px)
                        theta =Angle(tPA[i], 'deg') + 90*ut.deg

                        reg = EllipsePixelRegion(center=ce, width=a, height=b, angle=theta)
                        ellipse = reg.as_artist(facecolor='none', edgecolor='magenta', lw=2)
                        patch1.append(ellipse)
                        #kwar = dict(arrowprops=dict(arrowstyle="->", ec=".5",
                        #                              relpos=(0.5, 0.5)),
                        #              bbox=dict(boxstyle="round", ec="none", fc="w"))
                        ax1.annotate(i+1, xy=(x,y),
                                    xytext=(0, 0), textcoords="offset points", color="magenta")
                                    #ha="right", va="top")#, **kwar)
                        fetch_q.otext.append({f'S_TGSS-{i+1}': f'{s_tgss.tolist()[i]} {str(s_tgss.unit)}'})
                        fetch_q.otext.append({f'S_TGSS_e-{i+1}': f'{np.round(es_tgss.tolist()[i],3)} {str(es_tgss.unit)}'})
                    tgss_catalog = PatchCollection(
                    patch1, edgecolor='magenta', facecolor='None')
                    
                    ax1.add_collection(tgss_catalog)
                
                finally:
                    patch2 = []
                    for i in range(len(ncen[0])):
                        x, y = ncen[0][i], ncen[1][i]
                        ce = PixCoord(x, y)
                        a = to_pixel(nmaj[i], fetch_q.r, px)
                        b = to_pixel(nmin[i], fetch_q.r, px)
                        if nPA[i] != 0 and nPA[i] != '--':
                            theta = Angle(nPA[i], 'deg') + 90*ut.deg
                        else:
                            theta = 0*ut.deg + 90*ut.deg

                        reg = EllipsePixelRegion(
                            center=ce, width=a, height=b, angle=theta)
                        ellipse = reg.as_artist(
                            facecolor='none', edgecolor='cyan', lw=2)
                        patch2.append(ellipse)
                        ax1.annotate(i+1, xy=(x, y),
                                    xytext=(0, 0), textcoords="offset points", color="cyan",
                                        ha="right", va="top")#, **kwar)
                        
                        fetch_q.otext.append({f'S_NVSS-{i+1}': f'{s_nvss.tolist()[i]} {str(s_nvss.unit)}'})
                        fetch_q.otext.append({f'S_NVSS_e-{i+1}': f'{np.round(es_nvss.tolist()[i],3)} {str(es_nvss.unit)}'})
                    nvss_catalog = PatchCollection(
                        patch2, edgecolor='cyan', facecolor='None')
                    ax1.add_collection(nvss_catalog)                
            except:
                fetch_q.info = "catalog data missing"
            finally:
                #ax1.legend(framealpha=0.0, labelcolor='white')
                if spidx_file is not None:
                    kwargs = dict(arrowprops=dict(arrowstyle="->", ec=".5",
                                    relpos=(0.5, 0.5)),
                    bbox=dict(boxstyle="round", ec="none", fc="w"))
                    xi, yi, spidx = find_spidx(spidx_file, fetch_q.c, fetch_q.r)
                    for ien in range(len(xi)):
                        #print(ien)
                        Xi, Yi = fetch_q.wcs.world_to_pixel(SkyCoord(xi[ien]*ut.deg, yi[ien]*ut.deg)) 
                        #print(Xi,Yi, spidx)
                        ax1.annotate(f'{spidx[ien]}', xy=(Xi, Yi),
                                        xytext=(1, -40), textcoords="offset points",
                                        ha="right", va="top", **kwargs)

                #-------- single survey plot ---------#--#
                dss2r = sqrt(dss2r, scale_min=np.percentile(
                    np.unique(dss2r), 1.), scale_max=np.percentile(np.unique(dss2r), 100.))
                ax2 = fig.add_subplot(1, 2, 2, projection=fetch_q.wcs)
                pl_RGB(ax2, dss2r, title='TGSS(GMRT)-NVSS(VLA)-FIRST(VLA)-DSS2R(DSS)', name=name, annot=annot)
                
                ax2.contour(nvss, lvlcn, colors='cyan')
                ax2.contour(tgss, lvlct, colors='magenta')
                ax2.contour(first, lvlcf, colors='yellow')

                leg1 = mpatches.Patch(color='cyan', label='NVSS')
                leg2 = mpatches.Patch(color='magenta', label='TGSS')
                leg3 = mpatches.Patch(color='yellow', label='FIRST')
                leg4 = mpatches.Patch(color='white', label='DSS2R')
                
                ax2.legend(handles=[leg1, leg2, leg3, leg4],
                        labelcolor='linecolor', framealpha=0.0,)
                ax2.autoscale(False)
                plt.subplots_adjust(wspace=0.01, hspace=0.01)

                #-------- Saving final plot ------#--#
                string1 = save_fig(plt, fig, kind)
                

                #-------- Output for success -----#--#
                fetch_q.uri.append(
                    {'img1': 'data:image/png;base64,' + urllib.parse.quote(string1)})
                time_taken = perf_counter()-start
                fetch_q.info = 'completed in ' + str(np.round(time_taken, 3))+". "
                fetch_q.status = "success"

        return fetch_q.throw_output()